/* This file is autogenerated */
#ifndef XEN_API_XEN_ALL_H
#define XEN_API_XEN_ALL_H
#include <xen/api/xen_common.h>
#include <xen/api/xen_console.h>
#include <xen/api/xen_console_protocol.h>
#include <xen/api/xen_crashdump.h>
#include <xen/api/xen_event.h>
#include <xen/api/xen_event_operation.h>
#include <xen/api/xen_host_cpu.h>
#include <xen/api/xen_host.h>
#include <xen/api/xen_host_metrics.h>
#include <xen/api/xen_int_float_map.h>
#include <xen/api/xen_int_int_map.h>
#include <xen/api/xen_int_string_set_map.h>
#include <xen/api/xen_network.h>
#include <xen/api/xen_on_crash_behaviour.h>
#include <xen/api/xen_on_normal_exit.h>
#include <xen/api/xen_pbd.h>
#include <xen/api/xen_pif.h>
#include <xen/api/xen_pif_metrics.h>
#include <xen/api/xen_sr.h>
#include <xen/api/xen_string_set.h>
#include <xen/api/xen_string_string_map.h>
#include <xen/api/xen_user.h>
#include <xen/api/xen_vbd.h>
#include <xen/api/xen_vbd_metrics.h>
#include <xen/api/xen_vbd_mode.h>
#include <xen/api/xen_vbd_type.h>
#include <xen/api/xen_vdi.h>
#include <xen/api/xen_vdi_type.h>
#include <xen/api/xen_vif.h>
#include <xen/api/xen_vif_metrics.h>
#include <xen/api/xen_vm_guest_metrics.h>
#include <xen/api/xen_vm.h>
#include <xen/api/xen_vm_metrics.h>
#include <xen/api/xen_vm_power_state.h>
#include <xen/api/xen_xspolicy.h>
#include <xen/api/xen_cpu_pool.h>
#endif
